//
//
//  Matrix.h
//
//      AUTHOR: Russell Lowke
//
//		  Date: May 12th 2005
//

#ifndef MATRIX_H
#define MATRIX_H

#include <iostream>
#include <cmath>
using namespace std;

// Forward declare class template and function template
template <class T> class Matrix;
template <class T> ostream& operator<<(ostream& os, Matrix<T>& v);

template <class T>
class Matrix {

	//friend ostream& operator<< (ostream& os, const Matrix<T>& m);

    public:
        
		Matrix<T>(size_t w = 0, size_t h = 0);					// default constructor
		Matrix<T>(const Matrix<T>& m);							// copy constructor
        Matrix<T>(size_t w, size_t h, T data[]);				// matrix of width and height with data
        virtual ~Matrix<T>();									// destructor

		void	   print();										// print matrix data

	    // comparisons and assignment
		bool       equal      (const Matrix<T>& other) const;
		Matrix<T>& operator=  (const Matrix<T>& other);
		bool       operator== (const Matrix<T>& other) const;

		// Matrix operations
		Matrix<T>& add        (const Matrix<T>& other);
		Matrix<T>& subtract   (const Matrix<T>& other);
		Matrix<T>  mMultiply  (const Matrix<T>& other) const;
		Matrix<T>& multiply   (const Matrix<T>& other);
		Matrix<T>& divide	  (const Matrix<T>& other);
        Matrix<T>& operator+= (const Matrix<T>& other);
		Matrix<T>  operator+  (const Matrix<T>& other) const;
        Matrix<T>& operator-= (const Matrix<T>& other);
		Matrix<T>  operator-  (const Matrix<T>& other) const;
	    Matrix<T>& operator*= (const Matrix<T>& other);
		Matrix<T>  operator*  (const Matrix<T>& other) const;
        Matrix<T>& operator/= (const Matrix<T>& other);
		Matrix<T>  operator/  (const Matrix<T>& other) const;

		// by value operations
		Matrix<T>& add		  (T value);
		Matrix<T>& subtract   (T value);
		Matrix<T>& multiply   (T value);
		Matrix<T>& divide     (T value);
		Matrix<T>& operator+= (T value);
		Matrix<T>  operator+  (T value) const;
		Matrix<T>& operator-= (T value);
		Matrix<T>  operator-  (T value) const;
		Matrix<T>& operator*= (T value);
		Matrix<T>  operator*  (T value) const;
		Matrix<T>& operator/= (T value);
		Matrix<T>  operator/  (T value) const;

		//  accessors
		inline T  operator()(size_t x, size_t y) const		{ return  m_data[x][y]; }
		inline T  getAt(size_t x, size_t y) const			{ return  m_data[x][y]; }
		inline size_t getWidth() const						{ return  m_width; }
		inline size_t getHeight() const						{ return  m_height; }
		inline T** getData() const							{ return  m_data; }

		// mutators
		inline void   setAt(size_t x, size_t y, T val)		{ m_data[x][y] = val; }

	protected:

		void initialize(size_t w, size_t h, T* data);
		void initialize(size_t w, size_t h, T** data);

		T** m_data;

    private:
    
		void initialize(size_t w, size_t h);
		inline bool congruent(const Matrix<T>& other) const { 
			return (m_width == other.m_width && m_height == other.m_height); 
		}

        size_t	m_width, m_height;
};

// default constructor
template <class T>
Matrix<T>::Matrix(size_t w /* = 0 */, size_t h /* = 0 */) {
	initialize(w, h);
}

// copy constructor
template <class T>
Matrix<T>::Matrix(const Matrix<T>& m) {				
	*this = m;
}

// general constructors
template <class T>
Matrix<T>::Matrix(size_t w, size_t h, T data[]) {
	initialize(w, h, data);
}

// destructor
template <class T>
Matrix<T>::~Matrix() {
	// delete two dimensional array
	for (size_t i = 0; i < m_width; ++i)
		delete[] m_data[i];
	delete[] m_data;
}

//
// initialize Matrix with data
//
template <class T>
void Matrix<T>::initialize(size_t w, size_t h) {
	m_width  = w;
    m_height = h;

	// c style two dimensional array (quick)
	m_data = new T* [m_width];
	for (size_t i = 0; i < m_width; ++i)
		m_data[i] = new T[m_height];
}
template <class T>
void Matrix<T>::initialize(size_t w, size_t h, T* data) {
	initialize(w, h);
	for (size_t x = 0; x < m_width; ++x)
		for (size_t y = 0; y < m_height; ++y)
			m_data[x][y] = data[ m_width * y + x ];
}
template <class T>
void Matrix<T>::initialize(size_t w, size_t h, T** data) {
	initialize(w, h);
	for (size_t x = 0; x < m_width; ++x)
		for (size_t y = 0; y < m_height; ++y )
			m_data[x][y] = data[x][y];
}

template <class T>
void Matrix<T>::print() {
    for (size_t y = 1; y < m_height; ++y) {
        cout << "[ ";
        for (size_t x = 0; x < m_width; ++x)
            cout << m_data[x][y] << " ";
        cout << "]" << endl;
    }
}

//
// comparison & assignment operators
//
template <class T>
bool Matrix<T>::equal(const Matrix<T>& other) const {
	if (! congruent(other)) return false;
	for (size_t x = 0; x < m_width; ++x)				// test data
		for (size_t y = 0; y < m_height; ++y)
			if (m_data[x][y] != other.m_data[x][y])
				return false;
	return true;										// ok.
}
template <class T>
Matrix<T>& Matrix<T>::operator=(const Matrix<T>& other) {
	if (this == &other) return *this;
	initialize(other.m_width, other.m_height, other.m_data);
	return *this;
}
template <class T>
bool Matrix<T>::operator==(const Matrix<T>& other) const {
	return equal(other);
}

//
//	matrix operations
//
template <class T>
Matrix<T>& Matrix<T>::add( const Matrix<T>& other) {
    if (congruent(other))
		for (size_t x = 0; x < m_width; ++x)
			for (size_t y = 0; y < m_height; ++y )
				m_data[x][y] += other.m_data[x][y];
    return *this;
}
template <class T>
Matrix<T>& Matrix<T>::subtract(const Matrix<T>& other) {
    if (congruent(other))
		for (size_t x = 0; x < m_width; ++x)
			for (size_t y = 0; y < m_height; ++y )
				m_data[x][y] -= other.m_data[x][y];
    return *this;
}
//
// for matrix multiplication the number of columns
// in this Matrix must equal the number of rows in other
template <class T>
Matrix<T> Matrix<T>::mMultiply(const Matrix<T>& other) const {
	size_t height = m_height;
	size_t width  = other.m_width;	
	Matrix<T> result(width, height);
	T n = 0;
	for (size_t x = 0; x < width; ++x)
		for (size_t y = 0; y < height; result.m_data[x][y] = n, ++y, n = 0 )
			for (size_t i = 0; i < other.m_height && i < m_width; ++i )
				n += (m_data[i][y] * other.m_data[x][i]);
    return result;
}
template <class T>
Matrix<T>& Matrix<T>::multiply(const Matrix<T>& other) {
    if (congruent(other))
		for (size_t x = 0; x < m_width; ++x)
			for (size_t y = 0; y < m_height; ++y )
				m_data[x][y] *= other.m_data[x][y];
    return *this;
}
template <class T>
Matrix<T>& Matrix<T>::divide( const Matrix<T>& other) {
    if (congruent(other))
		for (size_t x = 0; x < m_width; ++x)
			for (size_t y = 0; y < m_height; ++y )
				m_data[x][y] /= other.m_data[x][y];
    return *this;
}
template <class T>
Matrix<T>& Matrix<T>::operator+= (const Matrix<T>& other) {
    return add(other);
}
template <class T>
Matrix<T> Matrix<T>::operator+ (const Matrix<T>& other) const {
	Matrix<T> temp = *this;
	temp += other;
	return temp;
}
template <class T>
Matrix<T>& Matrix<T>::operator-= (const Matrix<T>& other) {
    return subtract(other);
}
template <class T>
Matrix<T> Matrix<T>::operator- (const Matrix<T>& other) const {
	Matrix<T> temp = *this;
	temp -= other;
	return temp;
}
template <class T>
Matrix<T>& Matrix<T>::operator*= (const Matrix<T>& other) {
	if (m_width == other.m_height) {
		*this = mMultiply(other);
		return *this;
	} else
		return multiply(other);
}
template <class T>
Matrix<T> Matrix<T>::operator* (const Matrix<T>& other) const {
	if (m_width == other.m_height)
		return mMultiply(other);
	else {
		Matrix<T> temp = *this;
		temp *= other;
		return temp;
	}
}
template <class T>
Matrix<T>& Matrix<T>::operator/= (const Matrix<T>& other) {
    return divide(other);
}
template <class T>
Matrix<T> Matrix<T>::operator/ (const Matrix<T>& other) const {
	Matrix<T> temp = *this;
	temp /= other;
	return temp;
}

//
// by value operations
//
template <class T>
Matrix<T>& Matrix<T>::add(T value) {
	for (size_t x = 0; x < m_width; ++x)
		for (size_t y = 0; y < m_height; ++y )
			m_data[x][y] += value;
    return *this;
}
template <class T>
Matrix<T>& Matrix<T>::subtract(T value) {
	for (size_t x = 0; x < m_width; ++x)
		for (size_t y = 0; y < m_height; ++y )
			m_data[x][y] -= value;
    return *this;
}
template <class T>
Matrix<T>& Matrix<T>::multiply(T value) {
	for (size_t x = 0; x < m_width; ++x)
		for (size_t y = 0; y < m_height; ++y )
			m_data[x][y] *= value;
    return *this;
}
template <class T>
Matrix<T>& Matrix<T>::divide(T value) {
	for (size_t x = 0; x < m_width; ++x)
		for (size_t y = 0; y < m_height; ++y )
			m_data[x][y] /= value;
    return *this;
}
template <class T>
Matrix<T>& Matrix<T>::operator+= (T value) {
    return add(value);
}
template <class T>
Matrix<T>  Matrix<T>::operator+ (T value) const {
	Matrix<T> temp = *this;
	temp += value;
	return temp;
}
template <class T>
Matrix<T>& Matrix<T>::operator-= (T value) {
    return subtract(value);
}
template <class T>
Matrix<T>  Matrix<T>::operator- (T value) const {
	Matrix<T> temp = *this;
	temp -= value;
	return temp;
}
template <class T>
Matrix<T>& Matrix<T>::operator*= (T value) {
    return multiply(value);
}
template <class T>
Matrix<T>  Matrix<T>::operator* (T value) const {
	Matrix<T> temp = *this;
	temp *= value;
	return temp;
}
template <class T>
Matrix<T>& Matrix<T>::operator/= (T value) {
    return divide(value);
}
template <class T>
Matrix<T>  Matrix<T>::operator/ (T value) const {
	Matrix<T> temp = *this;
	temp /= value;
	return temp;
}


//
//	print Matrix
//
template <class T>
ostream& operator<< (ostream& os, const Matrix<T>& m) {
    for (size_t y = 0; y < m.m_height; ++y) {
        os << "[ ";
        for (size_t x = 0; x < m.m_width; ++x)
            os << m.m_data[x][y] << " ";
        os << "]" << endl;
    }
	return os;
}


class Point2D : public Matrix<double> {

	public:
		Point2D(double x = 0, double y = 0) {
			double data[] = { x, y };
			initialize(1, 2, data);
		}
		Point2D(const Matrix<double>& m) { initialize(m.getWidth(), m.getHeight(), m.getData()); }

		inline double distance(const Point2D& other) const {
			return sqrt(pow(other.x() - m_data[0][0], 2) + pow(other.y() - m_data[0][1], 2));
		}

		inline double  x() const				{ return m_data[0][0]; }
		inline double  y() const				{ return m_data[0][1]; }
		inline void   x(const double n)		{ m_data[0][0] = n; }
		inline void   y(const double n)		{ m_data[0][1] = n; }
		
};

class Point3D : public Matrix<double> {

	public:
		Point3D(double x = 0, double y = 0, double z = 0) {
			double data[] = { x, y, z };
			initialize(1, 3, data);
		}
		Point3D(const Matrix<double>& m) { initialize(m.getWidth(), m.getHeight(), m.getData()); }

		inline double distance(const Point3D& other) const {
			return sqrt(pow(other.x() - m_data[0][0], 2) + pow(other.y() - m_data[0][1], 2) + pow(other.z() - m_data[0][2], 2));
		}

		inline double  x() const				{ return m_data[0][0]; }
		inline double  y() const				{ return m_data[0][1]; }
		inline double  z() const				{ return m_data[0][2]; }
		inline void   x(const double n)		{ m_data[0][0] = n; }
		inline void   y(const double n)		{ m_data[0][1] = n; }
		inline void   z(const double n)		{ m_data[0][2] = n; }
		
};

class Point4D : public Matrix<double> {

	public:
		Point4D(double x = 0, double y = 0, double z = 0, double w = 0) {
			double data[] = { x, y, z, w };
			initialize(1, 4, data);
		}
		Point4D(const Matrix<double>& m) { initialize(m.getWidth(), m.getHeight(), m.getData()); }

		inline double distance(const Point4D& other) const {
			return sqrt(pow(other.x() - m_data[0][0], 2) + pow(other.y() - m_data[0][1], 2) + pow(other.z() - m_data[0][2], 2));
		}

		inline double  x() const				{ return m_data[0][0]; }
		inline double  y() const				{ return m_data[0][1]; }
		inline double  z() const				{ return m_data[0][2]; }
		inline double  w() const				{ return m_data[0][3]; }
		inline void   x(const double n)		{ m_data[0][0] = n; }
		inline void   y(const double n)		{ m_data[0][1] = n; }
		inline void   z(const double n)		{ m_data[0][2] = n; }
		inline void   w(const double n)		{ m_data[0][3] = n; }

};

class Array : public Matrix<double> {

	public:
		Array(int n) {
			double data[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };
			initialize(1, n+1, data);
		}

		inline double operator()(size_t n) const      { return  m_data[0][n]; }
		inline void   set(size_t n, const double val) { m_data[0][n] = val; }
};


class Array2D : public Matrix<double> {

	public:
		Array2D(int w, int h) {
			double data[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 
							 10,11,12,13,14,15,16,17,18,19,
							 20,21,22,23,24,25,26,27,28,29,
							 30,31,32,33,34,35,36,37,38,39,
							 40,41,42,43,44,45,46,47,48,49,
							 50,51,52,53,54,55,56,57,58,59,
							 60,61,62,63,64,65,66,67,68,69,
							 70,71,72,73,74,75,76,77,78,79,
							 80,81,82,83,84,85,86,87,88,89,
							 90,91,92,93,94,95,96,97,98,99, 100};

			initialize(w+1, h+1, data);
		}
};



class Rect : public Matrix<double> {

	public:
		Rect(double x1 = 0, double y1 = 0, double x2 = 0, double y2 = 0) {
			double data[] = { x1, y1, x2, y2 };
			initialize(1, 4, data);
		}
		Rect(const Matrix<double>& m) { initialize(m.getWidth(), m.getHeight(), m.getData()); }

		inline Point2D loc() const { 
			return Point2D(m_data[0][0] + (m_data[0][2]-m_data[0][0])/2, m_data[0][1] + (m_data[0][3]-m_data[0][1])/2); 
		}
		inline bool inside(const Point2D& pt) const {
			return (inside(pt.x(), pt.y()));
		}
		inline bool inside(const double& x, const double& y) const {
			return (x >= ((m_data[0][0] < m_data[0][2]) ? m_data[0][0] : m_data[0][2]) && 
				    x <= ((m_data[0][0] > m_data[0][2]) ? m_data[0][0] : m_data[0][2]) && 
					y >= ((m_data[0][1] < m_data[0][3]) ? m_data[0][1] : m_data[0][3]) && 
					y <= ((m_data[0][1] > m_data[0][3]) ? m_data[0][1] : m_data[0][3]));
		}
		inline bool intersect(const Rect& other) const {
			return (other.inside(m_data[0][0], m_data[0][1]) || 
				    other.inside(m_data[0][0], m_data[0][3]) ||
					other.inside(m_data[0][2], m_data[0][1]) ||
					other.inside(m_data[0][2], m_data[0][3]));
		}

		inline double  x1() const			{ return m_data[0][0]; }
		inline double  y1() const			{ return m_data[0][1]; }
		inline double  x2() const			{ return m_data[0][2]; }
		inline double  y2() const			{ return m_data[0][3]; }
		inline void   x1(const double n)		{ m_data[0][0] = n; }
		inline void   y1(const double n)		{ m_data[0][1] = n; }
		inline void   x2(const double n)		{ m_data[0][2] = n; }
		inline void   y2(const double n)		{ m_data[0][3] = n; }
		
};

#endif /* MATRIX_H */